@@ -26,6 +26,7 @@ constexpr double DEFAULT_SCALING         = 1.0;
 constexpr int DEFAULT_INTERP             = 1;
 constexpr double DEFAULT_FREQ            = 0.0;
 const uhd::rfnoc::io_type_t DEFAULT_TYPE = uhd::rfnoc::IO_TYPE_SC16;
+constexpr double TX_SIGN                 = -1.0;
 
 //! Space (in bytes) between register banks per channel
 constexpr uint32_t REG_CHAN_OFFSET = 2048;
@@ -103,6 +104,7 @@ public:
             _register_props(chan);
         }
         register_issue_stream_cmd();
+        register_issue_tune_request();
     }
 
     double set_freq(const double freq,
@@ -443,6 +445,121 @@ private:
         post_action(dst_edge, new_action);
     }
 
+    void register_issue_tune_request()
+    {
+        RFNOC_LOG_TRACE("DUC register_isssue_tune_request");
+        register_action_handler(ACTION_KEY_TUNE_REQUEST,
+            [this](const res_source_info& src, action_info::sptr action) {
+                tune_request_action_info::sptr tune_request_action =
+                    std::dynamic_pointer_cast<tune_request_action_info>(action);
+                if (!tune_request_action) {
+                    throw uhd::runtime_error(
+                        "Received tune_request of invalid action type!");
+                }
+                issue_tune_request_action_handler(src, tune_request_action);
+            });
+    }
+
+    void issue_tune_request_action_handler(
+        const res_source_info& src, tune_request_action_info::sptr tune_request_action)
+    {
+        RFNOC_LOG_TRACE("DUC isssue_tune_request_action_handler");
+        res_source_info dst_edge{res_source_info::invert_edge(src.type), src.instance};
+        const size_t chan                = src.instance;
+        uhd::tune_request_t tune_request = tune_request_action->tune_request;
+
+        RFNOC_LOG_TRACE("Received tune_request to "
+                        << src.to_string() << ", id==" << tune_request_action->id);
+
+        uhd::freq_range_t dsp_range    = get_frequency_range(chan);
+        uhd::freq_range_t tune_range   = tune_request_action->overall_freq_range;
+        tune_request_action->dsp_range = dsp_range;
+
+        if (tune_range.empty()) {
+            tune_range = dsp_range;
+        }
+
+        if (src.type == res_source_info::OUTPUT_EDGE) {
+            auto set_dsp_freq = [this, chan](
+                                    double freq) { set_freq(freq, chan, boost::none); };
+
+            double clipped_requested_freq = tune_range.clip(tune_request.target_freq);
+            tune_request_action->tune_result.target_dsp_freq = abs(
+                tune_request_action->tune_result.actual_rf_freq - clipped_requested_freq);
+
+            //------------------------------------------------------------------
+            //-- Set the DSP frequency depending upon the DSP frequency policy.
+            //------------------------------------------------------------------
+            double target_dsp_freq = 0.0;
+            switch (tune_request.dsp_freq_policy) {
+                case uhd::tune_request_t::POLICY_AUTO:
+                    // invert the sign on the dsp freq for transmit (spinning up vs down)
+                    tune_request_action->tune_result.target_dsp_freq *= TX_SIGN;
+
+                    /* If we are using the AUTO tuning policy, then we prevent the
+                     * CORDIC from spinning us outside of the range of the baseband
+                     * filter, regardless of what the user requested. This could happen
+                     * if the user requested a center frequency so far outside of the
+                     * tunable range of the FE that the CORDIC would spin outside the
+                     * filtered baseband. */
+                    target_dsp_freq = tune_request_action->tune_result.target_dsp_freq
+                                      - tune_request_action->tune_result.actual_dsp_freq;
+                    break;
+
+                case uhd::tune_request_t::POLICY_MANUAL:
+                    /* If the user has specified a manual tune policy, we will allow
+                     * tuning outside of the baseband filter, but will still clip the
+                     * target DSP frequency to within the bounds of the CORDIC to
+                     * prevent undefined behavior (likely an overflow). */
+                    target_dsp_freq = dsp_range.clip(tune_request.dsp_freq);
+                    break;
+
+                case uhd::tune_request_t::POLICY_NONE:
+                    break; // does not set
+            }
+            RFNOC_LOG_TRACE(
+                str(boost::format("Target DSP Freq: %.6fMHz") % (target_dsp_freq / 1e6)));
+
+            //------------------------------------------------------------------
+            //-- Tune the DSP
+            //------------------------------------------------------------------
+            if (tune_request.dsp_freq_policy != uhd::tune_request_t::POLICY_NONE) {
+                set_dsp_freq(target_dsp_freq);
+            }
+            const double actual_dsp_freq = get_freq(chan);
+
+            RFNOC_LOG_TRACE(
+                str(boost::format("Actual DSP Freq: %.6fMHz") % (actual_dsp_freq / 1e6)));
+
+            tune_request_action->tune_result.target_dsp_freq = target_dsp_freq;
+            tune_request_action->tune_result.actual_dsp_freq += actual_dsp_freq;
+
+            RFNOC_LOG_TRACE(
+                "Tune_result details DUC : target_rf_frq = "
+                << tune_request_action->tune_result.target_rf_freq
+                << " target dsp freq = "
+                << tune_request_action->tune_result.target_dsp_freq << " clipped rf_freq "
+                << tune_request_action->tune_result.clipped_rf_freq
+                << " actual_rf_freq = " << tune_request_action->tune_result.actual_rf_freq
+                << " actual dsp_freq= "
+                << tune_request_action->tune_result.actual_dsp_freq);
+
+            auto new_action                = tune_request_action_info::make(tune_request);
+            new_action->tune_request       = tune_request_action->tune_request;
+            new_action->tune_result        = tune_request_action->tune_result;
+            new_action->dsp_range          = tune_request_action->dsp_range;
+            new_action->overall_freq_range = tune_request_action->overall_freq_range;
+            post_action(dst_edge, new_action);
+        } else {
+            auto new_action                = tune_request_action_info::make(tune_request);
+            new_action->tune_request       = tune_request_action->tune_request;
+            new_action->tune_result        = tune_request_action->tune_result;
+            new_action->dsp_range          = tune_request_action->dsp_range;
+            new_action->overall_freq_range = tune_request_action->overall_freq_range;
+            post_action(dst_edge, new_action);
+        }
+    }
+
     /**************************************************************************
      * FPGA communication (register IO)
      *************************************************************************/
@@ -528,14 +645,10 @@ private:
     double _set_freq(
         const double requested_freq, const double dds_rate, const size_t chan)
     {
-        static int freq_word_width = 24;
         double actual_freq;
         int32_t freq_word;
         std::tie(actual_freq, freq_word) =
-            get_freq_and_freq_word(requested_freq, dds_rate, freq_word_width);
-
-        // Only the upper 24 bits of the SR_FREQ_ADDR register are used, so shift the word
-        freq_word <<= (32 - freq_word_width);
+            get_freq_and_freq_word(requested_freq, dds_rate);
 
         _duc_reg_iface.poke32(
             SR_FREQ_ADDR, uint32_t(freq_word), chan, get_command_time(chan));
